@@ -138,10 +138,19 @@
 
     Updated by      Richard Gooch   17-APR-2000: Added "COPY" action.
 
-    Last updated by Richard Gooch   18-APR-2000: Moved some things to header
+    Updated by      Richard Gooch   18-APR-2000: Moved some things to header
   "devfsd.h". Removed special argument "ENTRY" and renamed "INFO" to "EVENT"
   for "CFUNCTION" action. Bug fix in debug mode with shared objects.
 
+    Updated by      Richard Gooch   20-APR-2000: Changed version macro.
+
+    Updated by      Richard Gooch   24-APR-2000: Created a table of rules for
+  simple compatibility translations. Split part of <read_config> into
+  <process_config_line>, in preparation for NIS support.
+
+    Last updated by Richard Gooch   26-APR-2000: Added NIS configuration
+  support.
+
 
 */
 #include <unistd.h>
@@ -168,6 +177,8 @@
 #include <regex.h>
 #include <errno.h>
 #include <dlfcn.h>
+#include <rpcsvc/ypclnt.h>
+#include <rpcsvc/yp_prot.h>
 #include <karma.h>
 #include "devfsd.h"
 
@@ -180,7 +191,7 @@
 #endif
 
 
-#define VERSION "1.3.5"
+#define DEVFSD_VERSION "1.3.6"
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
 
@@ -287,8 +298,11 @@ flag syslog_is_open = FALSE;
 
 /*  Private functions  */
 static void setup_initial_entries ();
-static void read_config (CONST char *filename, flag optional,
+static void read_config (CONST char *location, flag optional,
 			 unsigned long *event_mask);
+static void process_config_line (CONST char *line, unsigned long *event_mask);
+static int process_yp_line (int instatus, char *inkey, int inkeylen,
+			    char *inval, int invallen, char *indata);
 static void do_servicing (int fd, unsigned long event_mask);
 static void service_name (const struct devfsd_notify_struct *info);
 static void action_permissions (const struct devfsd_notify_struct *info,
@@ -419,9 +433,9 @@ int main (int argc, char **argv)
 	 (DEVFSD_PROTOCOL_REVISION_DAEMON != proto_rev) )
     {
 	fprintf (stderr, "devfsd: Linux device filesystem daemon v%s\n",
-		 VERSION);
+		 DEVFSD_VERSION);
 	fprintf (stderr,
-		 "(C) 1998-1999  Richard Gooch <rgooch@atnf.csiro.au>\n\n");
+		 "(C) 1998-2000  Richard Gooch <rgooch@atnf.csiro.au>\n\n");
 	fprintf (stderr, "Daemon protocol revision:\t%d\n",
 		 DEVFSD_PROTOCOL_REVISION_DAEMON);
 	fprintf (stderr, "Kernel-side protocol revision:\t%d\n", proto_rev);
@@ -512,228 +526,262 @@ static void setup_initial_entries ()
 	symlink (curr->dest, curr->name);
 }   /*  End Function setup_initial_entries  */
 
-static void read_config (CONST char *filename, flag optional,
+static void read_config (CONST char *location, flag optional,
 			 unsigned long *event_mask)
-/*  [SUMMARY] Read the configuration file.
-    <filename> The filename to read from.
+/*  [SUMMARY] Read a configuration database.
+    <location> The location to read the database from. Either a filename or a
+    YP(NIS) map name prefixed with '+'.
     <optional> If TRUE, the routine will silently ignore a missing config file.
     <event_mask> The event mask is written here. This is not initialised.
     [RETURNS] Nothing.
 */
 {
-    char *ptr, *line;
     FILE *fp;
     char buf[STRING_LENGTH];
 
-    if ( ( fp = fopen (filename, "r") ) == NULL )
+    if (location[0] == '+')
+    {
+	/*  It's a YP map  */
+	struct ypall_callback ypall_cbk;
+
+	ypall_cbk.foreach = ( int (*) () ) process_yp_line;
+	ypall_cbk.data = (char *) event_mask;
+	if (yp_all (NULL, (char *) location + 1, &ypall_cbk) == 0)
+	{
+	    if (syslog_is_open)
+		SYSLOG (LOG_INFO, "read map: \"%s\"\n", location + 1);
+	    return;
+	}
+	if (optional) return;
+	SYSLOG (LOG_ERR, "error reading map: \"%s\"\n", location + 1);
+	exit (1);
+    }
+    if ( ( fp = fopen (location, "r") ) == NULL )
     {
 	if ( optional && (errno == ENOENT) ) return;
 	SYSLOG (LOG_ERR, "error opening file: \"%s\"\t%s\n",
-		CONFIG_FILE, ERRSTRING);
+		location, ERRSTRING);
 	exit (1);
     }
-    if (trace_level > 2) fprintf (stderr, "Reading file: \"%s\"\n",
-				  CONFIG_FILE);
+    if (trace_level > 2) fprintf (stderr, "Reading file: \"%s\"\n", location);
     while (fgets (buf, STRING_LENGTH, fp) != NULL)
     {
-	int err, num_args, count;
-	struct config_entry_struct *new;
-	char p[MAX_ARGS][STRING_LENGTH];
-	char when[STRING_LENGTH], what[STRING_LENGTH];
-	char name[STRING_LENGTH], tmp[STRING_LENGTH];
+	char *line;
 
 	buf[strlen (buf) - 1] = '\0';
 	/*  Skip whitespace  */
 	for (line = buf; isspace (*line); ++line);
 	if (line[0] == '\0') continue;
 	if (line[0] == '#') continue;
-	for (count = 0; count < MAX_ARGS; ++count) p[count][0] = '\0';
-	num_args = sscanf (line, "%s %s %s %s %s %s %s %s %s %s",
-			   when, name, what, p[0], p[1], p[2], p[3],
-			   p[4], p[5], p[6]);
-	if (strcasecmp (when, "CLEAR_CONFIG") == 0)
-	{
-	    free_config ();
-	    *event_mask = 0;
-	    continue;
-	}
-	if (num_args < 2)
-	{
-	    SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
-	    SYSLOG (LOG_ERR, "exiting\n");
-	    exit (1);
-	}
-	if ( (strcasecmp (when, "INCLUDE") == 0) ||
-	     (strcasecmp (when, "OPTIONAL_INCLUDE") == 0) )
-	{
-	    st_expr_expand (name, STRING_LENGTH, name, get_variable, NULL,
-			    NULL);
-	    read_config (name, (toupper (when[0]) == 'I') ? FALSE : TRUE,
-			 event_mask);
-	    continue;
-	}
-	if (num_args < 3)
-	{
-	    SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
-	    SYSLOG (LOG_ERR, "exiting\n");
-	    exit (1);
-	}
-	if ( ( new = malloc (sizeof *new) ) == NULL )
-	{
-	    SYSLOG (LOG_ERR, "error allocating\n");
-	    SYSLOG (LOG_ERR, "exiting\n");
-	    exit (1);
-	}
-	memset (new, 0, sizeof *new);
-	if (strcasecmp (when, "REGISTER") == 0)
-	    new->action.when = DEVFSD_NOTIFY_REGISTERED;
-	else if (strcasecmp (when, "UNREGISTER") == 0)
-	    new->action.when = DEVFSD_NOTIFY_UNREGISTERED;
-	else if (strcasecmp (when, "ASYNC_OPEN") == 0)
-	    new->action.when = DEVFSD_NOTIFY_ASYNC_OPEN;
-	else if (strcasecmp (when, "CLOSE") == 0)
-	    new->action.when = DEVFSD_NOTIFY_CLOSE;
-	else if (strcasecmp (when, "LOOKUP") == 0)
-	    new->action.when = DEVFSD_NOTIFY_LOOKUP;
-	else if (strcasecmp (when, "CHANGE") == 0)
-	    new->action.when = DEVFSD_NOTIFY_CHANGE;
-	else if (strcasecmp (when, "CREATE") == 0)
-	    new->action.when = DEVFSD_NOTIFY_CREATE;
-	else
-	{
-	    SYSLOG (LOG_ERR, "bad WHEN in config line: \"%s\"\n", line);
-	    SYSLOG (LOG_ERR, "exiting\n");
-	    exit (1);
-	}
-	if (strcasecmp (what, "PERMISSIONS") == 0)
-	{
-	    new->action.what = AC_PERMISSIONS;
-	    /*  Get user and group  */
-	    if ( ( ptr = strchr (p[0], '.') ) == NULL )
-	    {
-		SYSLOG (LOG_ERR, "missing '.' character in: \"%s\"\n", p[0]);
-		SYSLOG (LOG_ERR, "exiting\n");
-		exit (1);
-	    }
-	    *ptr++ = '\0';
-	    new->u.permissions.uid = get_uid (p[0]);
-	    new->u.permissions.gid = get_gid (ptr);
-	    /*  Get mode  */
-	    new->u.permissions.mode = get_mode (p[1]);
-	}
-	else if (strcasecmp (what, "MODLOAD") == 0)
-	    new->action.what = AC_MODLOAD;
-	else if (strcasecmp (what, "EXECUTE") == 0)
-	{
-	    new->action.what = AC_EXECUTE;
-	    num_args -= 3;
-	    if (strstr (p[0], "modprobe") != NULL)  /*  Bitch and moan  */
-		SYSLOG (LOG_NOTICE,
-			"PLEASE USE MODLOAD ACTION INSTEAD OF EXECUTE modprobe\n");
-	    for (count = 0; count < num_args; ++count)
-	    {
-		if ( ( new->u.execute.argv[count] =strdup (p[count]) ) ==NULL )
-		{
-		    SYSLOG (LOG_ERR, "error allocating\n");
-		    SYSLOG (LOG_ERR, "exiting\n");
-		    exit (1);
-		}
-	    }
-	    new->u.execute.argv[num_args] = NULL;
-	}
-	else if ( (strcasecmp (what, "MFUNCTION") == 0) ||
-		  (strcasecmp (what, "CFUNCTION") == 0) )
-	{
-	    new->action.what = (strcasecmp (what, "MFUNCTION") == 0) ?
-		AC_MFUNCTION : AC_CFUNCTION;
-	    num_args -= 3;
-	    if (num_args < 2)
-	    {
-		SYSLOG (LOG_ERR, "missing path and function in \"%s\"\n",line);
-		exit (1);
-	    }
-	    if ( ( new->u.function.so = get_shared_object (p[0]) ) == NULL )
-	    {
-		SYSLOG ( LOG_ERR, "error loading: \"%s\"\t%s\n",
-			 p[0], dlerror () );
-		exit (1);
-	    }
-	    if ( ( new->u.function.func.m =
-		   dlsym (new->u.function.so->handle, p[1]) ) == NULL )
-	    {
-		SYSLOG (LOG_ERR, "symbol: \"%s\" not found in %s\n",
-			p[1], p[0]);
-		exit (1);
-	    }
-	    --num_args;
-	    for (count = 0; count < num_args; ++count)
-	    {
-		if ( ( new->u.function.argv[count] = strdup (p[count + 1]) )
-		     == NULL )
-		{
-		    SYSLOG (LOG_ERR, "error allocating\n");
-		    SYSLOG (LOG_ERR, "exiting\n");
-		    exit (1);
-		}
-	    }
-	    new->u.function.argv[num_args] = NULL;
-	}
-	else if (strcasecmp (what, "COPY") == 0)
-	{
-	    new->action.what = AC_COPY;
-	    num_args -= 3;
-	    if (num_args != 2)
-	    {
-		SYSLOG (LOG_ERR, "missing path and function in \"%s\"\n",line);
-		exit (1);
-	    }
-	    if ( ( new->u.copy.source = strdup (p[0]) ) == NULL )
-	    {
-		SYSLOG (LOG_ERR, "error allocating source string\n");
-		SYSLOG (LOG_ERR, "exiting\n");
-		exit (1);
-	    }
-	    if ( ( new->u.copy.destination = strdup (p[1]) ) == NULL )
-	    {
-		SYSLOG (LOG_ERR, "error allocating destination string\n");
-		SYSLOG (LOG_ERR, "exiting\n");
-		exit (1);
-	    }
-	}
-	else if (strcasecmp (what, "IGNORE") == 0)
-	    new->action.what = AC_IGNORE;
-	else if (strcasecmp (what, "MKOLDCOMPAT") == 0)
-	    new->action.what = AC_MKOLDCOMPAT;
-	else if (strcasecmp (what, "MKNEWCOMPAT") == 0)
-	    new->action.what = AC_MKNEWCOMPAT;
-	else if (strcasecmp (what, "RMOLDCOMPAT") == 0)
-	    new->action.what = AC_RMOLDCOMPAT;
-	else if (strcasecmp (what, "RMNEWCOMPAT") == 0)
-	    new->action.what = AC_RMNEWCOMPAT;
-	else
-	{
-	    SYSLOG (LOG_ERR, "bad WHAT in config line: \"%s\"\n", line);
-	    SYSLOG (LOG_ERR, "exiting\n");
-	    exit (1);
-	}
-	if ( ( err = regcomp (&new->preg, name, 0) ) != 0 )
-	{
-	    regerror (err, &new->preg, tmp, STRING_LENGTH);
-	    SYSLOG (LOG_ERR, "error compiling regexp: \"%s\"\t%s\n", name,tmp);
-	    SYSLOG (LOG_ERR, "exiting\n");
-	    exit (1);
-	}
-	*event_mask |= 1 << new->action.when;
-	new->next = NULL;
-	if (first_config == NULL) first_config = new;
-	else last_config->next = new;
-	last_config = new;
+	process_config_line (line, event_mask);
     }
     fclose (fp);
     if (syslog_is_open)
-	SYSLOG (LOG_INFO, "read config file: \"%s\"\n", CONFIG_FILE);
+	SYSLOG (LOG_INFO, "read config file: \"%s\"\n", location);
 }   /*  End Function read_config   */
 
+static void process_config_line (CONST char *line, unsigned long *event_mask)
+/*  [SUMMARY] Process a line from a configuration file.
+    <line> The configuration line.
+    <event_mask> The event mask is written here. This is not initialised.
+    [RETURNS] Nothing.
+*/
+{
+    int err, num_args, count;
+    struct config_entry_struct *new;
+    char p[MAX_ARGS][STRING_LENGTH];
+    char when[STRING_LENGTH], what[STRING_LENGTH];
+    char name[STRING_LENGTH], tmp[STRING_LENGTH];
+
+    for (count = 0; count < MAX_ARGS; ++count) p[count][0] = '\0';
+    num_args = sscanf (line, "%s %s %s %s %s %s %s %s %s %s",
+		       when, name, what,
+		       p[0], p[1], p[2], p[3], p[4], p[5], p[6]);
+    if (strcasecmp (when, "CLEAR_CONFIG") == 0)
+    {
+	free_config ();
+	*event_mask = 0;
+	return;
+    }
+    if (num_args < 2)
+    {
+	SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
+	SYSLOG (LOG_ERR, "exiting\n");
+	exit (1);
+    }
+    if ( (strcasecmp (when, "INCLUDE") == 0) ||
+	 (strcasecmp (when, "OPTIONAL_INCLUDE") == 0) )
+    {
+	st_expr_expand (name, STRING_LENGTH, name, get_variable, NULL,
+			NULL);
+	read_config (name, (toupper (when[0]) == 'I') ? FALSE : TRUE,
+		     event_mask);
+	return;
+    }
+    if (num_args < 3)
+    {
+	SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
+	SYSLOG (LOG_ERR, "exiting\n");
+	exit (1);
+    }
+    if ( ( new = malloc (sizeof *new) ) == NULL )
+    {
+	SYSLOG (LOG_ERR, "error allocating\n");
+	SYSLOG (LOG_ERR, "exiting\n");
+	exit (1);
+    }
+    memset (new, 0, sizeof *new);
+    if (strcasecmp (when, "REGISTER") == 0)
+	new->action.when = DEVFSD_NOTIFY_REGISTERED;
+    else if (strcasecmp (when, "UNREGISTER") == 0)
+	new->action.when = DEVFSD_NOTIFY_UNREGISTERED;
+    else if (strcasecmp (when, "ASYNC_OPEN") == 0)
+	new->action.when = DEVFSD_NOTIFY_ASYNC_OPEN;
+    else if (strcasecmp (when, "CLOSE") == 0)
+	new->action.when = DEVFSD_NOTIFY_CLOSE;
+    else if (strcasecmp (when, "LOOKUP") == 0)
+	new->action.when = DEVFSD_NOTIFY_LOOKUP;
+    else if (strcasecmp (when, "CHANGE") == 0)
+	new->action.when = DEVFSD_NOTIFY_CHANGE;
+    else if (strcasecmp (when, "CREATE") == 0)
+	new->action.when = DEVFSD_NOTIFY_CREATE;
+    else
+    {
+	SYSLOG (LOG_ERR, "bad WHEN in config line: \"%s\"\n", line);
+	SYSLOG (LOG_ERR, "exiting\n");
+	exit (1);
+    }
+    if (strcasecmp (what, "PERMISSIONS") == 0)
+    {
+	char *ptr;
+
+	new->action.what = AC_PERMISSIONS;
+	/*  Get user and group  */
+	if ( ( ptr = strchr (p[0], '.') ) == NULL )
+	{
+	    SYSLOG (LOG_ERR, "missing '.' character in: \"%s\"\n", p[0]);
+	    SYSLOG (LOG_ERR, "exiting\n");
+	    exit (1);
+	}
+	*ptr++ = '\0';
+	new->u.permissions.uid = get_uid (p[0]);
+	new->u.permissions.gid = get_gid (ptr);
+	/*  Get mode  */
+	new->u.permissions.mode = get_mode (p[1]);
+    }
+    else if (strcasecmp (what, "MODLOAD") == 0) new->action.what = AC_MODLOAD;
+    else if (strcasecmp (what, "EXECUTE") == 0)
+    {
+	new->action.what = AC_EXECUTE;
+	num_args -= 3;
+	if (strstr (p[0], "modprobe") != NULL)  /*  Bitch and moan  */
+	    SYSLOG (LOG_NOTICE,
+		    "PLEASE USE MODLOAD ACTION INSTEAD OF EXECUTE modprobe\n");
+	for (count = 0; count < num_args; ++count)
+	{
+	    if ( ( new->u.execute.argv[count] = strdup (p[count]) ) == NULL )
+	    {
+		SYSLOG (LOG_ERR, "error allocating\n");
+		SYSLOG (LOG_ERR, "exiting\n");
+		exit (1);
+	    }
+	}
+	new->u.execute.argv[num_args] = NULL;
+    }
+    else if ( (strcasecmp (what, "MFUNCTION") == 0) ||
+	      (strcasecmp (what, "CFUNCTION") == 0) )
+    {
+	new->action.what = (strcasecmp (what, "MFUNCTION") == 0) ?
+	    AC_MFUNCTION : AC_CFUNCTION;
+	num_args -= 3;
+	if (num_args < 2)
+	{
+	    SYSLOG (LOG_ERR, "missing path and function in \"%s\"\n",line);
+	    exit (1);
+	}
+	if ( ( new->u.function.so = get_shared_object (p[0]) ) == NULL )
+	{
+	    SYSLOG (LOG_ERR, "error loading: \"%s\"\t%s\n", p[0], dlerror () );
+	    exit (1);
+	}
+	if ( ( new->u.function.func.m =
+	       dlsym (new->u.function.so->handle, p[1]) ) == NULL )
+	{
+	    SYSLOG (LOG_ERR, "symbol: \"%s\" not found in %s\n", p[1], p[0]);
+	    exit (1);
+	}
+	--num_args;
+	for (count = 0; count < num_args; ++count)
+	{
+	    if ( ( new->u.function.argv[count] = strdup (p[count + 1]) )
+		 == NULL )
+	    {
+		SYSLOG (LOG_ERR, "error allocating\n");
+		SYSLOG (LOG_ERR, "exiting\n");
+		exit (1);
+	    }
+	}
+	new->u.function.argv[num_args] = NULL;
+    }
+    else if (strcasecmp (what, "COPY") == 0)
+    {
+	new->action.what = AC_COPY;
+	num_args -= 3;
+	if (num_args != 2)
+	{
+	    SYSLOG (LOG_ERR, "missing path and function in \"%s\"\n", line);
+	    exit (1);
+	}
+	if ( ( new->u.copy.source = strdup (p[0]) ) == NULL )
+	{
+	    SYSLOG (LOG_ERR, "error allocating source string\n");
+	    SYSLOG (LOG_ERR, "exiting\n");
+	    exit (1);
+	}
+	if ( ( new->u.copy.destination = strdup (p[1]) ) == NULL )
+	{
+	    SYSLOG (LOG_ERR, "error allocating destination string\n");
+	    SYSLOG (LOG_ERR, "exiting\n");
+	    exit (1);
+	}
+    }
+    else if (strcasecmp (what, "IGNORE") == 0) new->action.what = AC_IGNORE;
+    else if (strcasecmp (what, "MKOLDCOMPAT") == 0)
+	new->action.what = AC_MKOLDCOMPAT;
+    else if (strcasecmp (what, "MKNEWCOMPAT") == 0)
+	new->action.what = AC_MKNEWCOMPAT;
+    else if (strcasecmp (what, "RMOLDCOMPAT") == 0)
+	new->action.what = AC_RMOLDCOMPAT;
+    else if (strcasecmp (what, "RMNEWCOMPAT") == 0)
+	new->action.what = AC_RMNEWCOMPAT;
+    else
+    {
+	SYSLOG (LOG_ERR, "bad WHAT in config line: \"%s\"\n", line);
+	SYSLOG (LOG_ERR, "exiting\n");
+	exit (1);
+    }
+    if ( ( err = regcomp (&new->preg, name, 0) ) != 0 )
+    {
+	regerror (err, &new->preg, tmp, STRING_LENGTH);
+	SYSLOG (LOG_ERR, "error compiling regexp: \"%s\"\t%s\n", name,tmp);
+	SYSLOG (LOG_ERR, "exiting\n");
+	exit (1);
+    }
+    *event_mask |= 1 << new->action.when;
+    new->next = NULL;
+    if (first_config == NULL) first_config = new;
+    else last_config->next = new;
+    last_config = new;
+}   /*  End Function process_config_line   */
+
+static int process_yp_line (int instatus, char *inkey, int inkeylen,
+			    char *inval, int invallen, char *indata)
+{
+    if (instatus != YP_TRUE) return (instatus);
+    if (invallen > 0) process_config_line (inval, (unsigned long *) indata);
+    return (0);
+}   /*  End Function process_yp_line  */
+
 static void do_servicing (int fd, unsigned long event_mask)
 /*  [SUMMARY] Service devfs changes until a signal is received.
     <fd> The open control file.
@@ -1162,6 +1210,35 @@ static void action_copy (const struct devfsd_notify_struct *info,
     SYSLOG (LOG_ERR, "error copying: \"%s\" to \"%s\"\n", source, destination);
 }   /*  End Function action_copy  */
 
+struct translate_struct
+{
+    char *match;    /*  The string to match to (up to length)                */
+    char *format;   /*  Format of output, "%s" takes data past match string,
+			NULL is effectively "%s" (just more efficient)       */
+};
+
+static struct translate_struct translate_table[] =
+{
+    {"sound/",     NULL},
+    {"printers/",  "lp%s"},
+    {"v4l/",       NULL},
+    {"parports/",  "parport%s"},
+    {"fb/",        "fb%s"},
+    {"netlink/",   NULL},
+    {"loop/",      "loop%s"},
+    {"floppy/",    "fd%s"},
+    {"rd/",        "ram%s"},
+    {"md/",        "md%s"},
+    {"cua/",       "cua%s"},
+    {"tts/",       "ttyS%s"},
+    {"ip2/",       "ip2%s"},
+    {"ttf/",       "ttyF%s"},
+    {"cuf/",       "cuf%s"},
+    {"vc/",        "tty%s"},
+    {"misc/",      NULL},
+    {NULL,         NULL}
+};
+
 static void action_compat (const struct devfsd_notify_struct *info,
 			   unsigned int action)
 /*  [SUMMARY] Process a compatibility request.
@@ -1173,6 +1250,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
     const char *compat_name = NULL;
     const char *dest_name = info->devname;
     char *ptr;
+    struct translate_struct *trans;
     char compat_buf[STRING_LENGTH];
     char dest_buf[STRING_LENGTH];
     static char function_name[] = "action_compat";
@@ -1182,32 +1260,27 @@ static void action_compat (const struct devfsd_notify_struct *info,
     {
       case AC_MKOLDCOMPAT:
       case AC_RMOLDCOMPAT:
-	if (strncmp (info->devname, "sound/", 6) == 0)
-	    compat_name = info->devname + 6;
-	else if (strncmp (info->devname, "printers/", 9) == 0)
+	for (trans = translate_table; trans->match != NULL; ++trans)
 	{
-	    sprintf (compat_buf, "lp%s", info->devname + 9);
-	    compat_name = compat_buf;
+	    size_t len = strlen (trans->match);
+
+	    if (strncmp (info->devname, trans->match, len) == 0)
+	    {
+		if (trans->format == NULL) compat_name = info->devname + len;
+		else
+		{
+		    sprintf (compat_buf, trans->format, info->devname + len);
+		    compat_name = compat_buf;
+		}
+		break;
+	    }
 	}
-	else if (strncmp (info->devname, "v4l/", 4) == 0)
-	    compat_name = info->devname + 4;
-	else if (strncmp (info->devname, "parports/", 9) == 0)
-	{
-	    sprintf (compat_buf, "parport%s", info->devname + 9);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "fb/", 3) == 0)
-	{
-	    sprintf (compat_buf, "fb%s", info->devname + 3);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "sbp/", 4) == 0)
+	if (compat_name != NULL) break;
+	if (strncmp (info->devname, "sbp/", 4) == 0)
 	{
 	    sprintf (compat_buf, "sbpcd%u", info->minor);
 	    compat_name = compat_buf;
 	}
-	else if (strncmp (info->devname, "netlink/", 8) == 0)
-	    compat_name = info->devname + 8;
 	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
 		  (strcmp (info->devname + info->namelen - 7,"generic") == 0) )
 	{
@@ -1225,11 +1298,6 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    compat_name = compat_buf;
 	    if (info->devname[info->namelen - 1] != 'n') ++compat_name;
 	}
-	else if (strncmp (info->devname, "loop/", 5) == 0)
-	{
-	    sprintf (compat_buf, "loop%s", info->devname + 5);
-	    compat_name = compat_buf;
-	}
 	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
 		  (strncmp (ptr = (strrchr (info->devname, '/') + 1), "mt", 2)
 		   == 0) )
@@ -1243,21 +1311,6 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sprintf (compat_buf, "sr%u", info->minor);
 	    compat_name = compat_buf;
 	}
-	else if (strncmp (info->devname, "floppy/", 7) == 0)
-	{
-	    sprintf (compat_buf, "fd%s", info->devname + 7);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "rd/", 3) == 0)
-	{
-	    sprintf (compat_buf, "ram%s", info->devname + 3);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "md/", 3) == 0)
-	{
-	    sprintf (compat_buf, "md%s", info->devname + 3);
-	    compat_name = compat_buf;
-	}
 	else if ( (strncmp (info->devname, "scsi/", 5) == 0) &&
 		  (strcmp (info->devname + info->namelen - 4, "disc") == 0) )
 	{
@@ -1413,36 +1466,6 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    if (compat_buf[3] == '0') compat_buf[3] = '\0';
 	    compat_name = compat_buf;
 	}
-	else if (strncmp (info->devname, "cua/", 4) == 0)
-	{
-	    sprintf (compat_buf, "cua%s", info->devname + 4);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "tts/", 4) == 0)
-	{
-	    sprintf (compat_buf, "ttyS%s", info->devname + 4);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "ip2/", 4) == 0)
-	{
-	    sprintf (compat_buf, "ip2%s", info->devname + 4);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "ttf/", 4) == 0)
-	{
-	    sprintf (compat_buf, "ttyF%s", info->devname + 4);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "cuf/", 4) == 0)
-	{
-	    sprintf (compat_buf, "cuf%s", info->devname + 4);
-	    compat_name = compat_buf;
-	}
-	else if (strncmp (info->devname, "vc/", 3) == 0)
-	{
-	    sprintf (compat_buf, "tty%s", info->devname + 3);
-	    compat_name = compat_buf;
-	}
 	else if (strncmp (info->devname, "pty/", 4) == 0)
 	{
 	    int index = atoi (info->devname + 5);
@@ -1468,8 +1491,6 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sprintf (compat_buf, "%sjs%d", type, index);
 	    compat_name = compat_buf;
 	}
-	else if (strncmp (info->devname, "misc/", 5) == 0)
-	    compat_name = info->devname + 5;
 	break;
       case AC_MKNEWCOMPAT:
       case AC_RMNEWCOMPAT:
@@ -1554,7 +1575,6 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sprintf (dest_buf, "../%s", info->devname);
 	    dest_name = dest_buf;
 	}
-
 	else if ( (strncmp (info->devname, "ide/", 4) == 0) &&
 		  (strcmp (info->devname + info->namelen - 4, "disc") == 0) )
 	{
